import traceback
from typing import Optional, Any

from bussdcc.device import DeviceProtocol
from bussdcc.context import ContextProtocol
from bussdcc import message


class Device(DeviceProtocol):
    kind: str = "device"
    id: str
    ctx: ContextProtocol | None
    online: bool

    def __init__(self, *, id: str, config: Optional[dict[str, Any]] = None):
        self.id = id
        self.config = config or {}
        self.ctx = None
        self.online = False

    def attach(self, ctx: ContextProtocol) -> None:
        self.ctx = ctx
        try:
            self.connect()
        except Exception as e:
            if self.ctx:
                self.ctx.emit(
                    message.DeviceFailed(
                        device=self.id,
                        kind=self.kind,
                        error=repr(e),
                        traceback=traceback.format_exc(),
                    )
                )
            raise
        self.online = True
        if self.ctx:
            self.ctx.emit(message.DeviceAttached(device=self.id, kind=self.kind))

    def detach(self) -> None:
        try:
            self.disconnect()
        except Exception as e:
            if self.ctx:
                self.ctx.emit(
                    message.DeviceFailed(
                        device=self.id,
                        kind=self.kind,
                        error=repr(e),
                        traceback=traceback.format_exc(),
                    )
                )
        finally:
            self.online = False
            if self.ctx:
                self.ctx.emit(message.DeviceDetached(device=self.id, kind=self.kind))
            self.ctx = None

    def connect(self) -> None:
        """Open hardware / resources"""
        pass

    def disconnect(self) -> None:
        """Close hardware / resources"""
        pass



from typing import Dict

from bussdcc.context import ContextProtocol
from bussdcc.process import ProcessProtocol
from bussdcc import message


class InterfaceManager:
    def __init__(self, ctx: ContextProtocol):
        self._ctx = ctx
        self._interfaces: Dict[str, ProcessProtocol] = {}

    def register(self, interface: ProcessProtocol) -> None:
        if interface.name in self._interfaces:
            raise ValueError(f"Interface `{interface.name}` already registered")

        self._interfaces[interface.name] = interface

    def list(self) -> list[ProcessProtocol]:
        return list(self._interfaces.values())

    def boot(self) -> None:
        for interface in self._interfaces.values():
            interface.attach(self._ctx)
            interface.start(self._ctx)
            self._ctx.emit(message.InterfaceStarted(interface=interface.name))

    def shutdown(self) -> None:
        for interface in self._interfaces.values():
            try:
                interface.stop(self._ctx)
            finally:
                interface.detach()
                self._ctx.emit(message.InterfaceStopped(interface=interface.name))



from .protocol import EventHandler, EventBusProtocol, SubscriptionProtocol
from .event import Event
from .bus import EventBus

__all__ = [
    "Event",
    "EventHandler",
    "EventBus",
    "EventBusProtocol",
    "SubscriptionProtocol",
]



from typing import Dict

from bussdcc.context import ContextProtocol
from bussdcc.process import ProcessProtocol
from bussdcc import message


class ProcessManager:
    def __init__(self, ctx: ContextProtocol):
        self._ctx = ctx
        self._processes: Dict[str, ProcessProtocol] = {}

    def register(self, process: ProcessProtocol) -> None:
        if process.name in self._processes:
            raise ValueError(f"Process `{process.name}` already registered")

        self._processes[process.name] = process

    def list(self) -> list[ProcessProtocol]:
        return list(self._processes.values())

    def boot(self) -> None:
        for process in self._processes.values():
            process.attach(self._ctx)
            process.start(self._ctx)
            self._ctx.emit(message.ProcessStarted(process=process.name))

    def shutdown(self) -> None:
        for process in self._processes.values():
            try:
                process.stop(self._ctx)
            finally:
                process.detach()
                self._ctx.emit(message.ProcessStopped(process=process.name))



from typing import Protocol, Callable, TypeVar

from .event import Event
from ..message import Message

T = TypeVar("T")

EventHandler = Callable[[Event[T]], None]


class SubscriptionProtocol(Protocol):
    def cancel(self) -> None: ...


class EventBusProtocol(Protocol):
    def emit(self, event: Event[Message]) -> None: ...
    def subscribe(
        self, event_type: type[T], handler: EventHandler[T]
    ) -> SubscriptionProtocol: ...
    def unsubscribe(self, subscription: SubscriptionProtocol) -> None: ...



from typing import Protocol, TYPE_CHECKING

from bussdcc.clock import ClockProtocol
from bussdcc.message import Message
from bussdcc.event.protocol import EventBusProtocol
from bussdcc.state.protocol import StateStoreProtocol

if TYPE_CHECKING:
    from bussdcc.runtime.protocol import RuntimeProtocol


class ContextProtocol(Protocol):
    clock: ClockProtocol
    runtime: "RuntimeProtocol"
    events: EventBusProtocol
    state: StateStoreProtocol

    def emit(self, message: Message) -> None: ...



from dataclasses import dataclass
from datetime import datetime
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(frozen=True)
class Event(Generic[T]):
    time: datetime
    payload: T



from typing import Callable, Generic, TypeVar, cast

from .event import Event
from ..message import Message

T = TypeVar("T")


class TypedHandler(Generic[T]):
    def __init__(
        self,
        event_type: type[T],
        handler: Callable[[Event[T]], None],
    ):
        self._event_type = event_type
        self._handler = handler

    def handle(self, event: Event[Message]) -> None:
        if isinstance(event.payload, self._event_type):
            self._handler(cast(Event[T], event))



import time
import threading
from datetime import datetime, timedelta

from bussdcc.clock import ClockProtocol


class ReplayClock(ClockProtocol):
    def __init__(self, *, speed: float = 1.0):
        self.speed = speed
        self._lock = threading.Condition()
        self._now: datetime | None = None
        self._monotonic = 0.0
        self._running = True

    def now_utc(self) -> datetime:
        assert self._now is not None, "Replay clock not started"
        return self._now

    def monotonic(self) -> float:
        return self._monotonic

    def sleep(self, seconds: float, cancel: threading.Event | None = None) -> bool:
        target = self._monotonic + seconds

        with self._lock:
            while self._running and self._monotonic < target:
                if cancel and cancel.is_set():
                    return True
                self._lock.wait()

        return False

    def advance_to(self, new_time: datetime) -> None:
        with self._lock:
            if self._now is None:
                self._now = new_time
                return

            delta = (new_time - self._now).total_seconds()
            if delta > 0 and self.speed > 0:
                time.sleep(delta / self.speed)

            self._now = new_time
            self._monotonic += max(delta, 0)

            self._lock.notify_all()

    def start(self, start_time: datetime) -> None:
        with self._lock:
            if self._now is not None:
                raise RuntimeError("ReplayClock already started")

            self._now = start_time
            self._monotonic = 0.0
            self._lock.notify_all()

    def stop(self) -> None:
        with self._lock:
            self._running = False
            self._lock.notify_all()



from typing import Protocol, Iterator

from bussdcc.context.protocol import ContextProtocol
from bussdcc.event import Event
from bussdcc.message import Message


class EventSourceProtocol(Protocol):
    def __iter__(self) -> Iterator[Event[Message]]: ...


class EventSinkProtocol(Protocol):
    def start(self, ctx: ContextProtocol) -> None: ...
    def stop(self) -> None: ...
    def handle(self, evt: Event[Message]) -> None: ...



from typing import Any, Callable
import threading

from .protocol import StateStoreProtocol
from .path import parse_path


class StateStore(StateStoreProtocol):
    """
    Thread-safe hierarchical state storage.

    Supports dot-separated paths:
        se = StateStore()
        se.set("system.clock.uptime", 42)
        se.get("system.clock.uptime") -> 42
    """

    def __init__(self) -> None:
        self._lock = threading.RLock()
        self._state: dict[str, Any] = {}

    def set(self, path: str, value: Any) -> None:
        with self._lock:
            keys = parse_path(path)
            d = self._state
            for k in keys[:-1]:
                nxt = d.setdefault(k, {})
                if not isinstance(nxt, dict):
                    raise TypeError(f"Path conflict at '{k}'")
                d = nxt
            d[keys[-1]] = value

    def get(self, path: str, default: Any = None) -> Any:
        with self._lock:
            keys = parse_path(path)
            d = self._state
            for k in keys:
                if not isinstance(d, dict) or k not in d:
                    return default
                d = d[k]
            return d

    def update(self, path: str, fn: Callable[[Any], Any]) -> Any:
        with self._lock:
            keys = parse_path(path)
            d = self._state

            for k in keys[:-1]:
                nxt = d.setdefault(k, {})
                if not isinstance(nxt, dict):
                    raise TypeError(f"Path conflict at '{k}'")
                d = nxt

            leaf = keys[-1]
            old = d.get(leaf)
            new = fn(old)
            d[leaf] = new
            return new



import threading
from datetime import datetime, timezone
import time

from .protocol import ClockProtocol


class SystemClock(ClockProtocol):
    def __init__(self) -> None:
        self._origin = time.monotonic()

    def now_utc(self) -> datetime:
        return datetime.now(timezone.utc)

    def monotonic(self) -> float:
        return time.monotonic()

    def uptime(self) -> float:
        return time.monotonic() - self._origin

    def sleep(
        self,
        seconds: float,
        cancel: threading.Event | None = None,
    ) -> bool:
        if cancel is None:
            time.sleep(seconds)
            return False

        deadline = self.monotonic() + seconds

        while True:
            remaining = deadline - self.monotonic()
            if remaining <= 0:
                return False  # normal wake

            # wait at most 0.5s so cancellation is responsive
            if cancel.wait(min(remaining, 0.5)):
                return True  # interrupted



from dataclasses import asdict, dataclass
from typing import ClassVar, Any

from .severity import Severity


@dataclass(slots=True, frozen=True)
class Message:
    severity: ClassVar[Severity] = Severity.INFO
    _registry: ClassVar[dict[str, type["Message"]]] = {}

    def __init_subclass__(cls) -> None:
        name = cls.__name__
        existing = cls._registry.get(name)

        if existing is None:
            cls._registry[name] = cls
            return

        if existing.__module__ == cls.__module__:
            cls._registry[name] = cls
            return

        raise RuntimeError(
            f"Duplicate message definition for {name}: " f"{existing} vs {cls}"
        )

    @classmethod
    def _resolve(cls, key: str) -> type["Message"]:
        try:
            return cls._registry[key]
        except KeyError:
            raise KeyError(f"Unknown message type: {key}") from None

    @classmethod
    def _key(cls) -> str:
        return cls.__name__

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)



from typing import Protocol
from datetime import datetime
import threading


class ClockProtocol(Protocol):
    def sleep(
        self,
        seconds: float,
        cancel: threading.Event | None = None,
    ) -> bool: ...
    def monotonic(self) -> float: ...
    def now_utc(self) -> datetime: ...



from dataclasses import dataclass
from typing import Optional

from .base import Message


@dataclass(slots=True, frozen=True)
class RuntimeBooting(Message):
    version: str


@dataclass(slots=True, frozen=True)
class RuntimeBooted(Message):
    version: str


@dataclass(slots=True, frozen=True)
class RuntimeShuttingDown(Message):
    reason: str | None


@dataclass(slots=True, frozen=True)
class RuntimeShutdown(Message):
    version: str



from dataclasses import dataclass

from bussdcc.event import Event

from .base import Message
from .severity import Severity


@dataclass(slots=True, frozen=True)
class InterfaceStarted(Message):
    interface: str


@dataclass(slots=True, frozen=True)
class InterfaceStopped(Message):
    interface: str



from typing import Optional
from datetime import datetime

from bussdcc.context import Context, ContextProtocol
from bussdcc.clock import ClockProtocol, ReplayClock
from bussdcc.event import EventBusProtocol
from bussdcc.state import StateStoreProtocol
from bussdcc.io import EventSourceProtocol

from .runtime import Runtime


class ReplayRuntime(Runtime):
    def __init__(
        self,
        *,
        clock: Optional[ClockProtocol] = None,
        events: EventBusProtocol | None = None,
        state: StateStoreProtocol | None = None,
        speed: float = 1.0,
        start_at: datetime | None = None,
    ):
        replay_clock = clock or ReplayClock(speed=speed)
        if not isinstance(replay_clock, ReplayClock):
            raise TypeError("ReplayRuntime requires ReplayClock")

        self._replay_clock = replay_clock
        self.speed = speed
        self.start_at = start_at
        super().__init__(clock=replay_clock, events=events, state=state)

    def replay(self, source: EventSourceProtocol) -> None:
        it = iter(source)

        try:
            first = next(it)
        except StopIteration:
            return

        start_time = self.start_at or first.time

        self._replay_clock.start(start_time)
        self.boot()

        # advance to first event if needed
        if first.time > start_time:
            self._replay_clock.advance_to(first.time)

        self.events.emit(first)

        for evt in source:
            if not self.booted:
                break
            if hasattr(self._replay_clock, "advance_to"):
                self._replay_clock.advance_to(evt.time)
            self.events.emit(evt)

        self._replay_clock.stop()

        self.shutdown("Replay Complete")



import threading
import traceback
from dataclasses import dataclass
from typing import Dict, List

from bussdcc.context.protocol import ContextProtocol
from bussdcc.service.protocol import ServiceProtocol
from bussdcc.message import service as message


@dataclass(frozen=True)
class ServiceInfo:
    name: str
    running: bool
    enabled: bool
    attached: bool
    interval: float
    restart: bool
    critical: bool


@dataclass
class _ServiceEntry:
    service: ServiceProtocol
    attached: bool = False
    running: bool = False
    thread: threading.Thread | None = None
    stop_event: threading.Event | None = None


class ServiceSupervisor:
    def __init__(self, ctx: ContextProtocol):
        self._ctx = ctx
        self._services: Dict[str, _ServiceEntry] = {}
        self._lock = threading.RLock()

    def register(self, service: ServiceProtocol) -> None:
        with self._lock:
            if service.name in self._services:
                raise ValueError(f"Service `{service.name}` already registered")

            self._services[service.name] = _ServiceEntry(service=service)

    def names(self) -> list[str]:
        with self._lock:
            return list(self._services)

    def list(self) -> list[ServiceProtocol]:
        with self._lock:
            return [entry.service for entry in self._services.values()]

    def get(self, name: str) -> ServiceProtocol | None:
        with self._lock:
            entry = self._services.get(name)
            return entry.service if entry else None

    def status(self, name: str) -> ServiceInfo:
        with self._lock:
            entry = self._services[name]

            return ServiceInfo(
                name=entry.service.name,
                running=entry.running,
                enabled=entry.service.enabled,
                attached=entry.attached,
                interval=entry.service.interval,
                restart=entry.service.restart,
                critical=entry.service.critical,
            )

    def statuses(self) -> List[ServiceInfo]:
        with self._lock:
            return [
                ServiceInfo(
                    name=e.service.name,
                    running=e.running,
                    enabled=e.service.enabled,
                    attached=e.attached,
                    interval=e.service.interval,
                    restart=e.service.restart,
                    critical=e.service.critical,
                )
                for e in self._services.values()
            ]

    def boot(self) -> None:
        with self._lock:
            for entry in self._services.values():
                if not entry.attached:
                    try:
                        entry.service.attach(self._ctx)
                        entry.attached = True
                    except Exception as e:
                        self._ctx.emit(
                            message.ServiceFailure(
                                service=entry.service.name,
                                error=repr(e),
                                traceback=traceback.format_exc(),
                            )
                        )

            for name in self._services:
                self._start_locked(name)

    def stop_all(self) -> None:
        with self._lock:
            names = list(self._services)
        for name in names:
            self.stop(name)

    def start(self, name: str) -> None:
        with self._lock:
            self._start_locked(name)

    def is_running(self, name: str) -> bool:
        with self._lock:
            entry = self._services.get(name)
            return bool(entry and entry.running)

    def running(self) -> List[ServiceProtocol]:
        with self._lock:
            return [entry.service for entry in self._services.values() if entry.running]

    def stop(self, name: str) -> None:
        with self._lock:
            entry = self._services[name]

            if not entry.running:
                return

            assert entry.stop_event is not None
            assert entry.thread is not None

            entry.stop_event.set()

        thread = entry.thread

        if thread:
            thread.join()

        entry.service.stop(self._ctx)

        with self._lock:
            entry.running = False
            entry.thread = None
            entry.stop_event = None

    def enable(self, name: str) -> None:
        with self._lock:
            entry = self._services[name]
            entry.service.enabled = True

    def disable(self, name: str) -> None:
        with self._lock:
            entry = self._services[name]
            entry.service.enabled = False
            running = entry.running

        if running:
            self.stop(name)

    def shutdown(self) -> None:
        self.stop_all()

        with self._lock:
            services = list(self._services.values())

        for entry in services:
            try:
                entry.service.detach()
            except Exception as e:
                self._ctx.emit(
                    message.ServiceError(
                        service=entry.service.name,
                        error=repr(e),
                        traceback=traceback.format_exc(),
                    )
                )

    def _start_locked(self, name: str) -> None:
        """Caller must hold self._lock."""
        entry = self._services[name]

        if not entry.service.enabled or entry.running:
            return

        stop_event = threading.Event()

        thread = threading.Thread(
            target=self._run_service,
            name=f"svc:{name}",
            args=(entry, stop_event),
            daemon=True,
        )

        entry.thread = thread
        entry.stop_event = stop_event
        entry.running = True

        thread.start()

        self._ctx.emit(message.ServiceStarted(service=name))

    def _run_service(
        self,
        entry: _ServiceEntry,
        stop_event: threading.Event,
    ) -> None:
        svc = entry.service
        ctx = self._ctx

        try:
            svc.start(ctx)

            while not stop_event.is_set():

                try:
                    svc.tick(ctx)

                except Exception as e:
                    ctx.emit(
                        message.ServiceError(
                            service=svc.name,
                            error=repr(e),
                            traceback=traceback.format_exc(),
                        )
                    )

                    if svc.critical:
                        ctx.emit(
                            message.ServiceFailure(
                                service=svc.name,
                                error=repr(e),
                                traceback=traceback.format_exc(),
                            )
                        )
                        break

                    if not svc.restart:
                        break

                    ctx.emit(message.ServiceRestart(service=svc.name))

                ctx.clock.sleep(svc.interval, cancel=stop_event)
        finally:
            with self._lock:
                entry.running = False

            ctx.emit(message.ServiceStopped(service=svc.name))



from typing import List, TypeVar, cast
import traceback
import threading

from .event import Event
from .handler import TypedHandler
from .protocol import EventHandler, SubscriptionProtocol, EventBusProtocol

from ..message import Message, Severity
from .. import message

T = TypeVar("T")


class _Subscription:
    def __init__(self, engine: "EventBus", handler: TypedHandler[T]):
        self._engine = engine
        self._handler = handler
        self._cancelled = False

    def cancel(self) -> None:
        if not self._cancelled:
            self._engine._unsubscribe(self)
            self._cancelled = True


class EventBus(EventBusProtocol):
    def __init__(self) -> None:
        self._lock = threading.RLock()
        self._subscriptions: List[_Subscription] = []

    def subscribe(
        self, event_type: type[T], handler: EventHandler[T]
    ) -> SubscriptionProtocol:
        wrapped = TypedHandler(event_type, handler)
        sub = _Subscription(self, wrapped)
        with self._lock:
            self._subscriptions.append(sub)
        return sub

    def _unsubscribe(self, subscription: _Subscription) -> None:
        with self._lock:
            try:
                self._subscriptions.remove(subscription)
            except ValueError:
                pass

    def unsubscribe(self, subscription: SubscriptionProtocol) -> None:
        subscription.cancel()

    def emit(self, evt: Event[Message]) -> None:
        with self._lock:
            subs = list(self._subscriptions)

        for sub in subs:
            try:
                sub._handler.handle(evt)
            except Exception as e:
                # Never recurse on error-level events
                if evt.payload.severity >= Severity.ERROR:
                    continue

                try:
                    error_evt = Event(
                        time=evt.time,
                        payload=cast(
                            Message,
                            message.EventSubscriberError(
                                event=evt.payload._key(),
                                handler=repr(sub._handler),
                                error=repr(e),
                                traceback=traceback.format_exc(),
                            ),
                        ),
                    )

                    self.emit(error_evt)  # SAFE: guarded by level check

                except Exception:
                    # Absolute last-resort safety
                    pass



from dataclasses import dataclass

from bussdcc.event import Event

from .base import Message
from .severity import Severity


@dataclass(slots=True, frozen=True)
class ServiceError(Message):
    severity = Severity.ERROR

    service: str
    error: str
    evt: Event[Message] | None = None
    traceback: str | None = None


@dataclass(slots=True, frozen=True)
class ServiceFailure(Message):
    severity = Severity.CRITICAL

    service: str
    error: str
    traceback: str | None = None


@dataclass(slots=True, frozen=True)
class ServiceRestart(Message):
    service: str


@dataclass(slots=True, frozen=True)
class ServiceStopped(Message):
    service: str


@dataclass(slots=True, frozen=True)
class ServiceStarted(Message):
    service: str



from dataclasses import dataclass

from bussdcc.event import Event

from .base import Message
from .severity import Severity


@dataclass(slots=True, frozen=True)
class DeviceAttached(Message):
    device: str
    kind: str


@dataclass(slots=True, frozen=True)
class DeviceDetached(Message):
    device: str
    kind: str


@dataclass(slots=True, frozen=True)
class DeviceFailed(Message):
    severity = Severity.ERROR

    device: str
    kind: str
    error: str
    traceback: str | None = None



from bussdcc.event import Event
from bussdcc.context import ContextProtocol
from bussdcc.message import Message

from .protocol import ProcessProtocol


class Process(ProcessProtocol):
    name = "unnamed"

    ctx: ContextProtocol | None

    def attach(self, ctx: ContextProtocol) -> None:
        self.ctx = ctx

    def detach(self) -> None:
        self.ctx = None

    def start(self, ctx: ContextProtocol) -> None:
        pass

    def stop(self, ctx: ContextProtocol) -> None:
        pass

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        pass



from bussdcc.event import Event
from bussdcc.context import ContextProtocol
from bussdcc.message import Message

from .protocol import ServiceProtocol


class Service(ServiceProtocol):
    name = "unnamed"

    interval = 1.0
    enabled = True
    restart = True
    critical = False

    ctx: ContextProtocol | None = None

    def attach(self, ctx: ContextProtocol) -> None:
        self.ctx = ctx

    def detach(self) -> None:
        self.ctx = None

    def start(self, ctx: ContextProtocol) -> None:
        """Called once when the service starts"""
        pass

    def tick(self, ctx: ContextProtocol) -> None:
        """Called repeatedly while the service is running"""
        pass

    def stop(self, ctx: ContextProtocol) -> None:
        """Called once when the service is stopping"""
        pass

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        pass



from typing import Dict

from bussdcc.device import DeviceProtocol
from bussdcc.context import ContextProtocol


class DeviceManager:
    def __init__(self, ctx: ContextProtocol):
        self._ctx = ctx
        self._devices: Dict[str, DeviceProtocol] = {}

    def attach(self, device: DeviceProtocol, *, booted: bool) -> None:
        if device.id in self._devices:
            raise ValueError(f"Device {device.id} already attached")

        self._devices[device.id] = device

        if booted:
            device.attach(self._ctx)

    def detach(self, id: str) -> None:
        device = self._devices.pop(id, None)
        if device:
            device.detach()

    def get(self, id: str) -> DeviceProtocol | None:
        return self._devices.get(id)

    def list(self, *, kind: str | None = None) -> list[DeviceProtocol]:
        if kind is None:
            return list(self._devices.values())

        return [d for d in self._devices.values() if d.kind == kind]

    def boot(self) -> None:
        for device in self._devices.values():
            device.attach(self._ctx)

    def shutdown(self) -> None:
        for device in reversed(list(self._devices.values())):
            device.detach()



import traceback
from typing import Optional, Dict, Self, Type, Literal, TypeVar
from types import TracebackType

from bussdcc.context import Context, ContextProtocol
from bussdcc.clock import ClockProtocol, SystemClock
from bussdcc.device import DeviceProtocol, DeviceManager
from bussdcc.event import Event, EventBus, EventBusProtocol
from bussdcc.message import Message, Severity
from bussdcc.state import StateStore, StateStoreProtocol
from bussdcc.service import ServiceProtocol, ServiceSupervisor
from bussdcc.process import ProcessProtocol, ProcessManager
from bussdcc.interface import InterfaceManager
from bussdcc.version import get_version

from bussdcc import message

from .protocol import RuntimeProtocol


class Runtime(RuntimeProtocol):
    def __init__(
        self,
        *,
        clock: Optional[ClockProtocol] = None,
        events: EventBusProtocol | None = None,
        state: StateStoreProtocol | None = None,
    ):
        self.clock: ClockProtocol = clock or SystemClock()
        self.events: EventBusProtocol = events or EventBus()
        self.state: StateStoreProtocol = state or StateStore()
        self.version: str = get_version()
        self.ctx: ContextProtocol = Context(
            clock=self.clock, runtime=self, events=self.events, state=self.state
        )

        self._booted: bool = False
        self.interfaces = InterfaceManager(self.ctx)
        self.processes = ProcessManager(self.ctx)
        self.devices = DeviceManager(self.ctx)
        self.services = ServiceSupervisor(self.ctx)

    def _on_boot(self) -> None:
        """Hook for subclasses to initialize resources."""
        return None

    def _on_shutdown(self, reason: Optional[str] = None) -> None:
        """Hook for subclasses to release resources."""
        return None

    def _dispatch_to_process(
        self, process: ProcessProtocol, evt: Event[Message]
    ) -> None:
        try:
            process.handle_event(self.ctx, evt)
        except Exception as e:
            if evt.payload.severity >= Severity.ERROR:
                return

            self.ctx.emit(
                message.ProcessError(
                    process=process.name,
                    error=repr(e),
                    evt=evt,
                    traceback=traceback.format_exc(),
                )
            )

    def _dispatch_to_service(
        self, service: ServiceProtocol, evt: Event[Message]
    ) -> None:
        try:
            service.handle_event(self.ctx, evt)
        except Exception as e:
            if evt.payload.severity >= Severity.ERROR:
                return

            self.ctx.emit(
                message.ServiceError(
                    service=service.name,
                    error=repr(e),
                    evt=evt,
                    traceback=traceback.format_exc(),
                )
            )

    def _dispatch(self, evt: Event[Message]) -> None:
        """
        Authoritative dispatch entrypoint.

        Called for every emitted Message.
        """

        # Processes
        for process in self.processes.list():
            self._dispatch_to_process(process, evt)

        # Interfaces
        for interface in self.interfaces.list():
            self._dispatch_to_process(interface, evt)

        # Services (event-driven side)
        for service in self.services.running():
            self._dispatch_to_service(service, evt)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} booted={self._booted}>"

    def __enter__(self) -> Self:
        self.boot()

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        reason = None
        if exc_type is not None:
            reason = f"{exc_type.__name__}: {exc_val}"
        self.shutdown(reason=reason)

        # Returning False allows exceptions to propagate normally
        return False

    @property
    def booted(self) -> bool:
        return self._booted

    def boot(self) -> None:
        if self._booted:
            return

        self._sub = self.events.subscribe(Message, self._dispatch)
        self.ctx.emit(message.RuntimeBooting(version=self.version))
        self._on_boot()
        self.devices.boot()
        self.processes.boot()
        self.interfaces.boot()
        self._booted = True
        self.ctx.emit(message.RuntimeBooted(version=self.version))
        self.services.boot()

    def shutdown(self, reason: Optional[str] = None) -> None:
        if not self._booted:
            return  # Be idempotent on shutdown to avoid exit crashes

        try:
            self.ctx.emit(message.RuntimeShuttingDown(reason=reason))
            self.services.shutdown()
            self.interfaces.shutdown()
            self.processes.shutdown()
            self.devices.shutdown()
            self._on_shutdown(reason)
            self.ctx.emit(message.RuntimeShutdown(version=self.version))
            self._sub.cancel()
        finally:
            # Ensure state is reset even if on_shutdown fails
            self._booted = False



from typing import TypeVar

from bussdcc.clock import ClockProtocol
from bussdcc.event import Event
from bussdcc.message import Message
from bussdcc.runtime.protocol import RuntimeProtocol
from bussdcc.event.protocol import EventBusProtocol
from bussdcc.state.protocol import StateStoreProtocol

from .protocol import ContextProtocol

T = TypeVar("T")


class Context(ContextProtocol):
    def __init__(
        self,
        clock: ClockProtocol,
        runtime: RuntimeProtocol,
        events: EventBusProtocol,
        state: StateStoreProtocol,
    ):
        self.clock: ClockProtocol = clock
        self.runtime: RuntimeProtocol = runtime
        self.events = events
        self.state = state

    def emit(self, message: Message) -> None:
        self.events.emit(Event(time=self.clock.now_utc(), payload=message))



from dataclasses import dataclass

from bussdcc.event import Event

from .base import Message
from .severity import Severity


@dataclass(slots=True, frozen=True)
class SystemReload(Message):
    pass


@dataclass(slots=True, frozen=True)
class SystemSignal(Message):
    signal: int
    action: str


@dataclass(slots=True, frozen=True)
class EventSubscriberError(Message):
    severity = Severity.ERROR

    event: str
    handler: str
    error: str
    traceback: str | None = None



from dataclasses import dataclass

from bussdcc.event import Event

from .base import Message
from .severity import Severity


@dataclass(slots=True, frozen=True)
class ProcessStarted(Message):
    process: str


@dataclass(slots=True, frozen=True)
class ProcessStopped(Message):
    process: str


@dataclass(slots=True, frozen=True)
class ProcessError(Message):
    severity = Severity.ERROR

    process: str
    error: str
    evt: Event[Message] | None = None
    traceback: str | None = None
